use std::{collections::HashSet, path::Path, str::FromStr};

use itertools::Itertools;

use crate::{input::read_into_vec, error::AdventError};

impl FromStr for ScratchGame {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let split = s.split(':').collect_vec();
        let id = split[0].split_whitespace().collect_tuple::<(&str, &str)>().ok_or(AdventError("Game Id not found.".into()))?.1.parse::<i32>()?;
        let split = split[1].split('|').collect_vec();
        let my_numbers: HashSet<i32> = split[0].trim().split_whitespace().map(|x| {
            x.parse::<i32>()
        }).collect::<Result<HashSet<_>, _>>()?;
        let winning_numbers: HashSet<i32> = split[1].trim().split_whitespace().map(|x| -> Result<i32, AdventError> {
            x.parse::<i32>().map_err(Into::into)
        }).collect::<Result<HashSet<_>, _>>()?;
        
        Ok(ScratchGame {
            id,
            my_numbers,
            winning_numbers
        })
    }
}

struct ScratchGame {
    id: i32,
    my_numbers: HashSet<i32>,
    winning_numbers: HashSet<i32>
}

pub fn one() -> Result<u32, AdventError> {
    let games = read_into_vec::<ScratchGame>(Path::new("resources/input4.txt"))?;
    Ok(games.into_iter().fold(Ok(0), |acc: Result<u64, AdventError>, game: ScratchGame| {
        let acc = acc?;
        let my_winning_numbers = game.my_numbers.intersection(&game.winning_numbers);
        Ok(acc + match my_winning_numbers.count() {
            0 => 0,
            x => (2 as u64).pow((x-1).try_into()?)
        })
    })? as u32)
}

pub fn two() -> Result<u32, AdventError> {
    let games = read_into_vec::<ScratchGame>(Path::new("resources/input4.txt"))?;
    let mut vec = Vec::<u32>::new();
    vec.resize(games.len(), 1);
    
    Ok(games.into_iter().fold(0, | acc, game | {
        let my_winning_numbers = game.my_numbers.intersection(&game.winning_numbers);
        let nr_free_tickets = my_winning_numbers.count() as i32;
        let vec_index: usize = (game.id - 1) as usize;
        for index in game.id..(game.id + nr_free_tickets ) {
            let nr_to_count = vec[vec_index];
            for _ in 0..nr_to_count {
                vec[index as usize] += 1;
            }       
        }
        acc + vec[vec_index]
    }))
}